"""
Strategies for locating contexts.  Allows for arbitrarily complex caching
and context differentiation strategies.

All context locators must implement a `get` method that returns an
`OrderedDict`-like object.
"""


from collections import OrderedDict
import threading


class DefaultContextLocator:
    """
    One-to-one mapping between contexts and trackers.  Every tracker will
    get a new context instance and it will always be returned by this
    locator.
    """

    def __init__(self):
        self.context = OrderedDict()

    def get(self):
        """Get a reference to the context."""
        return self.context


class ThreadLocalContextLocator:
    """
    Returns a different context depending on the thread that the locator
    was called from.  Thus, contexts can be isolated from one another
    on thread boundaries.

    Note that this makes use of `threading.local(),`  which is typically
    monkey-patched by alternative python concurrency frameworks (like `gevent`).

    Calls to `threading.local()` are delayed until first usage in order to
    give the third-party concurrency libraries an opportunity to monkey monkey
    patch it.
    """

    def __init__(self):
        self.thread_local_data = None

    def get(self):
        """Return a reference to a thread-specific context"""
        if not self.thread_local_data:
            self.thread_local_data = threading.local()

        if not hasattr(self.thread_local_data, 'context'):
            self.thread_local_data.context = OrderedDict()

        return self.thread_local_data.context



"""
Helper classes for backend tests
"""

from unittest import TestCase
from contextlib import contextmanager
import time
import os
import random
import string


class InMemoryBackend:
    """A backend that simply stores all events in memory"""

    def __init__(self):
        super().__init__()
        self.events = []

    def send(self, event):
        """Store the event in a list"""
        self.events.append(event)


class IntegrationTestCase(TestCase):
    """
    Tests the integration between a backend and any external systems
    it makes use of.
    """

    # This is equivalent to decorating all subclasses with attr('integration')
    # which allows us to selectively run integration tests.
    integration = 1


class PerformanceTestCase(TestCase):
    """
    Reads parameters from the following environment variables:

    * EVENT_TRACKING_PERF_EVENTS - Number of events to send to the backend
    * EVENT_TRACKING_PERF_PAYLOAD_SIZE - Approximate size (in bytes) of a
      payload field to include in every event.
    * EVENT_TRACKING_PERF_THRESHOLD_SECONDS - Fail the test if it takes
      longer than this number of seconds to save all of the events.
    """

    # This is equivalent to decorating all subclasses with attr('performance')
    performance = 1

    # pylint: disable=invalid-envvar-default
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_events = int(os.getenv('EVENT_TRACKING_PERF_EVENTS', 20000))
        self.payload_size = int(os.getenv('EVENT_TRACKING_PERF_PAYLOAD_SIZE', 600))
        self.random_payload = ''.join(random.choice(string.ascii_letters) for _ in range(self.payload_size))
        self.threshold = float(os.getenv('EVENT_TRACKING_PERF_THRESHOLD_SECONDS', 1.75))

    @contextmanager
    def assert_execution_time_less_than_threshold(self):
        """
        Times the execution of the block within the context and raises an
        `AssertionError` if it is longer than `self.threshold`
        """
        start_time = time.time()
        yield
        elapsed_time = time.time() - start_time

        print('')
        print(f'Elapsed Time: {elapsed_time} seconds')
        print(f'Threshold: {self.threshold} seconds')
        print(f'Number of Events: {self.num_events}')
        print(f'Payload Size: {self.payload_size} bytes')
        print('Events per second: {}'.format(self.num_events / elapsed_time))

        if self.threshold >= 0:
            self.assertLessEqual(elapsed_time, self.threshold)



"""
Runs invasive tests to ensure that the tracking system can communicate with
an actual MongoDB instance.
"""


from datetime import datetime
from uuid import uuid4

from pytz import UTC

from eventtracking.backends.mongodb import MongoBackend
from eventtracking.backends.tests import InMemoryBackend, IntegrationTestCase
from eventtracking.tracker import Tracker


class TestMongoIntegration(IntegrationTestCase):
    """
    Makes use of a real MongoDB instance to ensure the backend is wired up
    properly to the external system.  These tests require a mongodb instance
    to be running on localhost listening on the default port.
    """

    def setUp(self):
        super().setUp()
        self.database_name = 'test_eventtracking_' + str(uuid4())
        self.mongo_backend = MongoBackend(database=self.database_name)
        self.memory_backend = InMemoryBackend()
        self.tracker = Tracker({
            'mongo': self.mongo_backend,
            'mem': self.memory_backend
        })

    def tearDown(self):
        self.mongo_backend.connection.drop_database(self.database_name)
        super().tearDown()

    def test_sequential_events(self):
        now = datetime.now(UTC)
        for i in range(10):
            self.tracker.emit('org.test.user.login', {
                'username': 'tester',
                'user_id': 10,
                'email': 'tester@eventtracking.org',
                'sequence': i,
                'current_time': now
            })

        # Ensure MongoDB has finished writing out the events before we
        # run our query.
        self.mongo_backend.connection.fsync()

        mem_events = {}
        for event in self.memory_backend.events:
            mem_events[event['data']['sequence']] = event
        self.assertEqual(len(mem_events), 10)

        cursor = self.mongo_backend.collection.find()
        self.assertEqual(cursor.count(), 10)

        for event in cursor:
            mem_event = mem_events[event['data']['sequence']]
            mem_event['_id'] = event['_id']
            if self.are_results_equal(mem_event, event):
                del mem_events[event['data']['sequence']]
        self.assertEqual(len(mem_events), 0)

    def are_results_equal(self, left, right):
        """
        Ensure two events are equivalent.

        We use a bit of special logic here when comparing timestamps since
        MongoDB apparently only stores millisecond precision timestamps.
        Thus, when comparing we ignore the datetime.millisecond property.
        """
        for event in [left, right]:
            self.remove_microseconds_from_timestamps(event)

        return left == right

    def remove_microseconds_from_timestamps(self, event):
        """Truncate the microseconds from the event timestamp"""
        event['timestamp'] = event['timestamp'].replace(microsecond=0)
        event['data']['current_time'] = event['data']['current_time'].replace(microsecond=0)



"""Unit tests for the Mongo backend"""

from unittest import TestCase
from unittest.mock import patch
from unittest.mock import sentinel

from pymongo.errors import PyMongoError
from bson.errors import BSONError

from eventtracking.backends.mongodb import MongoBackend


class TestMongoBackend(TestCase):
    """Unit tests for the Mongo backend"""

    def setUp(self):
        super().setUp()
        self.mongo_patcher = patch('eventtracking.backends.mongodb.MongoClient')
        self.addCleanup(self.mongo_patcher.stop)
        self.mongo_patcher.start()

        self.backend = MongoBackend()

    def test_mongo_backend(self):
        events = [{'test': 1}, {'test': 2}]

        self.backend.send(events[0])
        self.backend.send(events[1])

        # Check if we inserted events into the database

        calls = self.backend.collection.insert.mock_calls

        self.assertEqual(len(calls), 2)

        # Unpack the arguments and check if the events were used
        # as the first argument to collection.insert

        def first_argument(call):
            """Extract the first argument from a `mock.call`"""
            _, args, _ = call
            return args[0]

        self.assertEqual(events[0], first_argument(calls[0]))
        self.assertEqual(events[1], first_argument(calls[1]))

    def test_authentication_settings(self):
        backend = MongoBackend(user=sentinel.user, password=sentinel.password)
        backend.database.authenticate.assert_called_once_with(sentinel.user, sentinel.password)

    def test_mongo_pymongo_insertion_error(self):
        self.backend.collection.insert.side_effect = PyMongoError

        self.backend.send({'test': 1})
        # Ensure this error is caught

    def test_mongo_bson_insertion_error(self):
        self.backend.collection.insert.side_effect = BSONError

        self.backend.send({'test': 1})
        # Ensure this error is caught



"""Test context locators."""


from unittest.mock import sentinel
from unittest import TestCase
import threading

from eventtracking import locator


class TestThreadLocalContextLocator(TestCase):
    """Test context locators."""

    def test_multithreaded_context(self):
        self.locator = locator.ThreadLocalContextLocator()  # pylint: disable=attribute-defined-outside-init

        # Events emitted from the parent thread should have this context
        context = {sentinel.context_key: sentinel.context_value}

        # Events emitted from the child thread should have this context
        thread_context = {sentinel.thread_key: sentinel.thread_value}

        # Set once the child thread has setup its context
        thread_in_context = threading.Event()
        # Set once the parent has emitted its event
        parent_sent_event = threading.Event()

        def worker():
            """A simulated child thread"""
            # Setup a context in this thread.  This should not inherit the parent thread context!
            self.locator.get()['child'] = thread_context

            # At this point both the parent and child threads have entered their contexts, but
            # the child thread should only "see" the context it setup.
            self.assertEqual(self.locator.get(), {'child': thread_context})

            # Notify the parent that the child has setup its context.  At this point both the
            # parent and the child have entered their own contexts.
            thread_in_context.set()
            # Wait for the parent to emit its event
            parent_sent_event.wait()

            del self.locator.get()['child']

            self.assertEqual(self.locator.get(), {})

        self.locator.get()['parent'] = context

        other_thread = threading.Thread(target=worker)
        other_thread.start()

        # Wait for the thread to setup its context.
        thread_in_context.wait()

        # At this point both the parent and child threads have entered their contexts, but
        # the parent thread should only "see" the context it setup.
        self.assertEqual(self.locator.get(), {'parent': context})

        # Notify the thread that it can send its event and exit
        parent_sent_event.set()
        other_thread.join()

        del self.locator.get()['parent']

        self.assertEqual(self.locator.get(), {})



"""Event tracker backend that saves events to a python logger."""


from datetime import datetime
from datetime import date
import logging
import json

from pytz import UTC

MAX_EVENT_SIZE = 1024  # 1 KB


class LoggerBackend:
    """
    Event tracker backend that uses a python logger.

    Events are logged to the INFO level as JSON strings.
    """

    def __init__(self, **kwargs):
        """
        Event tracker backend that uses a python logger.

        `name` is an identifier for the logger, which should have
            been configured using the default python mechanisms.
        """
        name = kwargs.get('name', None)
        self.max_event_size = kwargs.get('max_event_size', MAX_EVENT_SIZE)
        self.event_logger = logging.getLogger(name)
        level = kwargs.get('level', 'info')
        self.log = getattr(self.event_logger, level.lower())

    def send(self, event):
        """Send the event to the standard python logger"""
        event_str = json.dumps(event, cls=DateTimeJSONEncoder)

        # TODO: do something smarter than simply dropping the event on
        # the floor.
        if self.max_event_size is None or len(event_str) <= self.max_event_size:
            self.log(event_str)


class DateTimeJSONEncoder(json.JSONEncoder):
    """JSON encoder aware of datetime.datetime and datetime.date objects"""

    def default(self, obj):  # pylint: disable=arguments-renamed
        """
        Serialize datetime and date objects of iso format.

        datatime objects are converted to UTC.
        """

        if isinstance(obj, datetime):
            if obj.tzinfo is None:
                # Localize to UTC naive datetime objects
                obj = UTC.localize(obj)  # pylint: disable=no-value-for-parameter
            else:
                # Convert to UTC datetime objects from other timezones
                obj = obj.astimezone(UTC)
            return obj.isoformat()
        elif isinstance(obj, date):
            return obj.isoformat()

        return super().default(obj)



"""Custom exceptions that are raised by this package"""


class EventEmissionExit(Exception):
    """
    Raising this exception indicates that no further processing of the event should occur and it should be dropped.

    This should only be raised by processors.
    """


class NoTransformerImplemented(Exception):
    """
    Raise this exception when there is no transformer implemented
    for an event.
    """


class NoBackendEnabled(Exception):
    """
    Raise this exception when there is no backend enabled
    for an event.
    """



"""Tests various configuration settings for the django tracker"""


from unittest import TestCase

from unittest.mock import sentinel
from django.test.utils import override_settings

from eventtracking import tracker
from eventtracking.django.django_tracker import DjangoTracker, override_default_tracker


TEST_TRACKER_NAME = 'django.test.tracker'


class TestConfiguration(TestCase):
    """Tests various configuration settings for the django tracker"""

    def setUp(self):
        super().setUp()
        override_default_tracker()
        self.tracker = tracker.get_tracker()

    @override_settings(EVENT_TRACKING_BACKENDS={
        "fake": {
            'ENGINE': 'eventtracking.django.tests.test_configuration.TrivialFakeBackend'
        }
    })
    def test_configure(self):
        self.configure_tracker()
        fake_backend = self.tracker.get_backend('fake')
        self.assertTrue(isinstance(fake_backend, TrivialFakeBackend))

    def configure_tracker(self):
        """Reads the tracker configuration from the Django settings"""
        self.tracker = DjangoTracker()

    @override_settings(EVENT_TRACKING_BACKENDS={
        "no_engine": {
            'OPTIONS': {}
        }
    })
    def test_ignore_no_engine(self):
        with self.assertRaises(ValueError):
            self.configure_tracker()

    @override_settings(EVENT_TRACKING_BACKENDS={
        "empty_engine": {
            'ENGINE': ''
        }
    })
    def test_configure_empty_engine(self):
        self.assert_fails_to_configure_with_error()

    def assert_fails_to_configure_with_error(self, error=ValueError):
        """
        Attempts to read the tracker configuration from the Django settings
        and ensures that the tracker construction fails.
        """
        with self.assertRaises(error):
            self.configure_tracker()

    @override_settings(EVENT_TRACKING_BACKENDS={
        "invalid_package": {
            'ENGINE': 'foo.BarBackend'
        }
    })
    def test_configure_invalid_package(self):
        self.assert_fails_to_configure_with_error()

    @override_settings(EVENT_TRACKING_BACKENDS={
        "no_package_invalid_class": {
            'ENGINE': 'BarBackend'
        }
    })
    def test_configure_no_package_invalid_class(self):
        self.assert_fails_to_configure_with_error()

    @override_settings(EVENT_TRACKING_BACKENDS={
        "invalid_class": {
            'ENGINE': 'eventtracking.django.tests.test_configuration.BarBackend'
        }
    })
    def test_configure_invalid_class(self):
        self.assert_fails_to_configure_with_error()

    @override_settings(EVENT_TRACKING_BACKENDS={
        'with_options': {
            'ENGINE': 'eventtracking.django.tests.test_configuration.FakeBackendWithOptions',
            'OPTIONS': {
                'option': sentinel.option_value
            }
        }
    })
    def test_configure_engine_with_options(self):
        self.configure_tracker()
        self.assertEqual(self.tracker.get_backend('with_options').option, sentinel.option_value)

    @override_settings(EVENT_TRACKING_BACKENDS={
        'without_options': {
            'ENGINE': 'eventtracking.django.tests.test_configuration.FakeBackendWithOptions'
        }
    })
    def test_configure_engine_missing_options(self):
        self.configure_tracker()
        self.assertEqual(self.tracker.get_backend('without_options').option, None)

    @override_settings(EVENT_TRACKING_BACKENDS={
        'extra_options': {
            'ENGINE': 'eventtracking.django.tests.test_configuration.FakeBackendWithOptions',
            'OPTIONS': {
                'option': sentinel.option_value,
                'extra_option': sentinel.extra_option_value
            }
        }
    })
    def test_configure_engine_with_extra_options(self):
        self.configure_tracker()
        self.assertEqual(self.tracker.get_backend('extra_options').option, sentinel.option_value)

    @override_settings(EVENT_TRACKING_BACKENDS={
        'extra_options': {
            'ENGINE': 'eventtracking.django.tests.test_configuration.NotABackend'
        }
    })
    def test_configure_class_not_a_backend(self):
        self.assert_fails_to_configure_with_error()

    @override_settings(EVENT_TRACKING_BACKENDS={
        'outer_backend': {
            'ENGINE': 'eventtracking.django.tests.test_configuration.NestedBackend',
            'OPTIONS': {
                'backends': {
                    'inner_backend': {
                        'ENGINE': 'eventtracking.django.tests.test_configuration.FakeBackendWithOptions',
                        'OPTIONS': {
                            'option': sentinel.option_value,
                            'extra_option': sentinel.extra_option_value
                        }
                    },
                    'nested_backend': {
                        'ENGINE': 'eventtracking.django.tests.test_configuration.NestedBackend',
                        'OPTIONS': {
                            'backends': {
                                'trivial': {
                                    'ENGINE': 'eventtracking.django.tests.test_configuration.TrivialFakeBackend'
                                }
                            },
                            'processors': [
                                {'ENGINE': 'eventtracking.django.tests.test_configuration.NopProcessor'},
                            ]
                        }
                    }
                },
                'processors': [
                    {'ENGINE': 'eventtracking.django.tests.test_configuration.NopProcessor'},
                    {'ENGINE': 'eventtracking.django.tests.test_configuration.NopProcessor'},
                ]
            }
        }
    })
    def test_configure_nested_backends(self):
        self.configure_tracker()
        outer_backend = self.tracker.get_backend('outer_backend')
        self.assertEqual(len(outer_backend.backends), 2)

        inner_backend = outer_backend.backends['inner_backend']
        self.assertEqual(inner_backend.option, sentinel.option_value)

        self.assertEqual(len(outer_backend.processors), 2)
        self.assertTrue(isinstance(outer_backend.processors[0], NopProcessor))
        self.assertTrue(isinstance(outer_backend.processors[1], NopProcessor))

        nested_backend = outer_backend.backends['nested_backend']
        self.assertEqual(len(nested_backend.backends), 1)
        self.assertTrue(isinstance(nested_backend.backends['trivial'], TrivialFakeBackend))
        self.assertTrue(isinstance(nested_backend.processors[0], NopProcessor))

    @override_settings(EVENT_TRACKING_ENABLED=True)
    def test_overrides_default_tracker(self):
        override_default_tracker()
        self.assertTrue(isinstance(tracker.get_tracker(), DjangoTracker))

    @override_settings(EVENT_TRACKING_ENABLED=False)
    def test_leaves_default_tracker_alone(self):
        override_default_tracker()
        self.assertTrue(isinstance(tracker.get_tracker(), tracker.Tracker))

    @override_settings(EVENT_TRACKING_PROCESSORS=[
        {
            'ENGINE': 'eventtracking.django.tests.test_configuration.NopProcessor'
        }
    ])
    def test_single_processor(self):
        self.configure_tracker()
        self.assertEqual(len(self.tracker.processors), 1)
        self.assertTrue(isinstance(self.tracker.processors[0], NopProcessor))

    @override_settings(EVENT_TRACKING_PROCESSORS=[
        {
            'ENGINE': 'eventtracking.django.tests.test_configuration.ProcessorWithOptions',
            'OPTIONS': {
                'option': sentinel.option_value
            }
        }
    ])
    def test_processor_with_options(self):
        self.configure_tracker()
        self.assertEqual(len(self.tracker.processors), 1)
        self.assertTrue(isinstance(self.tracker.processors[0], ProcessorWithOptions))
        self.assertEqual(self.tracker.processors[0].option, sentinel.option_value)

    @override_settings(EVENT_TRACKING_PROCESSORS=[
        {}
    ])
    def test_missing_processor_engine(self):
        with self.assertRaises(ValueError):
            self.configure_tracker()

    @override_settings(EVENT_TRACKING_PROCESSORS=[
        {
            'ENGINE': 'eventtracking.django.tests.test_configuration.NopProcessor'
        },
        {
            'ENGINE': 'eventtracking.django.tests.test_configuration.NopProcessor'
        }
    ])
    def test_multiple_processor(self):
        self.configure_tracker()
        self.assertEqual(len(self.tracker.processors), 2)
        self.assertTrue(isinstance(self.tracker.processors[0], NopProcessor))
        self.assertTrue(isinstance(self.tracker.processors[1], NopProcessor))


class TrivialFakeBackend:
    """A trivial fake backend without any options"""

    def send(self, event):
        """Don't actually send the event anywhere"""


class NotABackend:
    """A class that is not a backend"""


class FakeBackendWithOptions(TrivialFakeBackend):
    """A trivial fake backend with options"""

    def __init__(self, **kwargs):
        super().__init__()
        self.option = kwargs.get('option', None)


class NopProcessor:
    """Changes every event"""

    def __call__(self, event):
        pass


class ProcessorWithOptions:
    """Takes in an argument"""

    def __init__(self, **kwargs):
        self. option = kwargs.get('option', None)

    def __call__(self, event):
        pass


class NestedBackend(TrivialFakeBackend):
    """Supports other backends as children"""

    def __init__(self, backends=None, processors=None, **_kwargs):
        self.backends = backends or {}
        self.processors = processors or []



"""MongoDB event tracker backend."""


import logging

import pymongo
from pymongo import MongoClient
from pymongo.errors import PyMongoError
from bson.errors import BSONError


log = logging.getLogger(__name__)


class MongoBackend:
    """Class for a MongoDB event tracker Backend"""

    def __init__(self, **kwargs):
        """
        Connect to a MongoDB.

        :Parameters:

          - `host`: hostname
          - `port`: port
          - `user`: collection username
          - `password`: collection user password
          - `database`: name of the database
          - `collection`: name of the collection
          - `extra`: parameters to pymongo.MongoClient not listed above

        """

        super().__init__()

        # Extract connection parameters from kwargs

        host = kwargs.get('host', 'localhost')
        port = kwargs.get('port', 27017)

        user = kwargs.get('user', '')
        password = kwargs.get('password', '')

        db_name = kwargs.get('database', 'eventtracking')
        collection_name = kwargs.get('collection', 'events')

        # Other mongo connection arguments
        extra = kwargs.get('extra', {})

        # By default disable write acknowledgments, reducing the time
        # blocking during an insert
        extra['w'] = extra.get('w', 0)

        # Make timezone aware by default
        extra['tz_aware'] = extra.get('tz_aware', True)

        # Connect to database and get collection

        self.connection = MongoClient(
            host=host,
            port=port,
            **extra
        )

        self.database = self.connection[db_name]

        if user or password:
            self.database.authenticate(user, password)

        self.collection = self.database[collection_name]

        self._create_indexes()

    def _create_indexes(self):
        """Ensures the proper fields are indexed"""
        # WARNING: The collection will be locked during the index
        # creation. If the collection has a large number of
        # documents in it, the operation can take a long time.

        # TODO: The creation of indexes can be moved to a Django
        # management command or equivalent. There is also an option to
        # run the indexing on the background, without locking.
        self.collection.ensure_index([('time', pymongo.DESCENDING)])
        self.collection.ensure_index('name')

    def send(self, event):
        """Insert the event in to the Mongo collection"""
        try:
            self.collection.insert(event, manipulate=False)
        except (PyMongoError, BSONError):
            # The event will be lost in case of a connection error or any error
            # that occurs when trying to insert the event into Mongo.
            # pymongo will re-connect/re-authenticate automatically
            # during the next event.
            msg = 'Error inserting to MongoDB event tracker backend'
            log.exception(msg)



"""
Runs performance tests to determine if a significant performance regression has
been introduced to the MongoBackend.
"""


from uuid import uuid4


from eventtracking.backends.mongodb import MongoBackend
from eventtracking.backends.tests import PerformanceTestCase
from eventtracking.tracker import Tracker


class TestBackendPerformance(PerformanceTestCase):
    """
    Makes use of real backend systems to see how long it takes for the backend
    to commit events to stable storage.
    """

    def setUp(self):
        super().setUp()
        self.database_name = 'perf_test_eventtracking_' + str(uuid4())
        self.mongo_backend = MongoBackend(database=self.database_name)
        self.tracker = Tracker({
            'mongo': self.mongo_backend
        })

    def tearDown(self):
        self.mongo_backend.connection.drop_database(self.database_name)
        super().tearDown()

    def test_sequential_events(self):
        with self.assert_execution_time_less_than_threshold():
            for i in range(self.num_events):
                self.tracker.emit('perf.event', {
                    'sequence': i,
                    'payload': self.random_payload
                })



"""
Track application events.  Supports persisting events to multiple backends.

Best Practices:

* It is recommended that event types are namespaced using dot notation to
  avoid naming collisions, similar to DNS names.  For example:
  org.edx.video.stop, edu.mit.audio.stop
* Avoid using event type names that may cause collisions.  The burden is
  on the analyst to decide whether your event is equivalent to another
  and should be grouped accordingly etc.
* Do not emit events that you don't own.  This could negatively impact
  the analysis of the event stream.  If you suspect your event is
  equivalent to another, say so in your documenation, and the analyst
  can decide whether or not to group them.
"""


from contextlib import contextmanager
from datetime import datetime
import logging

from pytz import UTC

from eventtracking.locator import DefaultContextLocator
from eventtracking.backends.routing import RoutingBackend

UNKNOWN_EVENT_TYPE = 'unknown'
DEFAULT_TRACKER_NAME = 'default'
TRACKERS = {}
LOG = logging.getLogger(__name__)


class Tracker:
    """
    Track application events.  Holds references to a set of backends that will
    be used to persist any events that are emitted.
    """
    def __init__(self, backends=None, context_locator=None, processors=None):
        self.routing_backend = RoutingBackend(backends=backends, processors=processors)
        self.context_locator = context_locator or DefaultContextLocator()

    @property
    def located_context(self):
        """
        The thread local context for this tracker.
        """
        return self.context_locator.get()

    def get_backend(self, name):
        """Gets the backend that was configured with `name`"""
        return self.backends[name]

    @property
    def processors(self):
        """The list of registered processors"""
        return self.routing_backend.processors

    @property
    def backends(self):
        """The dictionary of registered backends"""
        return self.routing_backend.backends

    def emit(self, name=None, data=None):
        """
        Emit an event annotated with the UTC time when this function was called.

        `name` is a unique identification string for an event that has
            already been registered.
        `data` is a dictionary mapping field names to the value to include in the event.
            Note that all values provided must be serializable.

        """
        event = {
            'name': name or UNKNOWN_EVENT_TYPE,
            'timestamp': datetime.now(UTC),
            'data': data or {},
            'context': self.resolve_context()
        }

        self.routing_backend.send(event)

    def resolve_context(self):
        """
        Create a new dictionary that corresponds to the union of all of the
        contexts that have been entered but not exited at this point.
        """
        merged = {}
        for context in self.located_context.values():
            merged.update(context)
        return merged

    def enter_context(self, name, ctx):
        """
        Enter a named context.  Any events emitted after calling this
        method will contain all of the key-value pairs included in `ctx`
        unless overridden by a context that is entered after this call.
        """
        self.located_context[name] = ctx

    def exit_context(self, name):
        """
        Exit a named context.  This will remove all key-value pairs
        associated with this context from any events emitted after it
        is removed.
        """
        del self.located_context[name]

    @contextmanager
    def context(self, name, ctx):
        """
        Execute the block with the given context applied.  This manager
        ensures that the context is removed even if an exception is raised
        within the context.
        """
        self.enter_context(name, ctx)
        try:
            yield
        finally:
            self.exit_context(name)


def register_tracker(tracker, name=DEFAULT_TRACKER_NAME):
    """
    Makes a tracker globally accessible.  Providing no `name` parameter
    allows you to register the global default tracker that will be used
    by subsequent calls to `tracker.emit`.
    """
    TRACKERS[name] = tracker


def get_tracker(name=DEFAULT_TRACKER_NAME):
    """
    Gets a named tracker.  Defaults to the default global tracker.  Raises
    a `KeyError` if no such tracker has been registered by previously calling
    `register_tracker`.
    """
    return TRACKERS[name]


def emit(name=None, data=None):
    """Calls `Tracker.emit` on the default global tracker"""
    return get_tracker().emit(name=name, data=data)



"""Filter out events whose names aren't on a pre-configured whitelist"""


from eventtracking.processors.exceptions import EventEmissionExit


class NameWhitelistProcessor:
    """

    Filter out events whose names aren't on a pre-configured whitelist.

    `whitelist` is an iterable collection containing event names that should be allowed to pass.
    """

    def __init__(self, whitelist=None, **_kwargs):
        try:
            if isinstance(whitelist, str):
                raise TypeError

            self.whitelist = frozenset(whitelist)
        except TypeError as error:
            raise TypeError(
                'The NameWhitelistProcessor must be passed a collection of allowed names '
                'using the "whitelist" parameter'
            ) from error

    def __call__(self, event):
        if event['name'] not in self.whitelist:
            raise EventEmissionExit()

        return event



"""
Test the RegexFilter processor.
"""
from unittest.mock import sentinel
import ddt
from django.test import TestCase

from eventtracking.processors.exceptions import EventEmissionExit
from eventtracking.exceptions import ImproperlyConfigured
from eventtracking.processors.regex_filter import RegexFilter, ALLOWLIST, BLOCKLIST


@ddt.ddt
class TestRegexFilterProcessor(TestCase):
    """
    Test the RegexFilter processor.
    """

    def setUp(self):
        super().setUp()
        self.sample_event = {
            'name': str(sentinel.name),
            'event_type': 'edx.test.event',
            'time': '2020-01-01T12:12:12.000000+00:00',
            'event': {
                'key': 'value'
            },
            'session': '0000'
        }

    def test_with_invalid_configurations(self):
        with self.assertRaises(ImproperlyConfigured):
            RegexFilter(filter_type='INVALID_TYPE', regular_expressions=[])

        with self.assertRaises(ImproperlyConfigured):
            RegexFilter(filter_type=ALLOWLIST, regular_expressions=['***'])

    @ddt.data(
        (ALLOWLIST, ['sentinel.name', 'any'], True),
        (ALLOWLIST, ['not_matching', 'NOT_MATCHING'], False),
        (ALLOWLIST, [''], True),
        (ALLOWLIST, [], False),
        (BLOCKLIST, ['sentinel.name', 'any'], False),
        (BLOCKLIST, ['not_matching', 'NOT_MATCHING'], True),
        (BLOCKLIST, [''], False),
        (BLOCKLIST, [], True),
    )
    @ddt.unpack
    def test_with_multiple_regex(self, filter_type, expressions, should_pass):
        regex_filter = RegexFilter(filter_type=filter_type, regular_expressions=expressions)

        if should_pass:
            self.assertEqual(regex_filter(self.sample_event), self.sample_event)
        else:
            with self.assertRaises(EventEmissionExit):
                regex_filter(self.sample_event)



"""
Tests for celery tasks.
"""
from unittest.mock import sentinel

from django.test import TestCase
from django.test.utils import override_settings

from eventtracking.tasks import send_event
from eventtracking.tracker import get_tracker
from eventtracking.django.django_tracker import override_default_tracker


MOCK_EVENT_TRACKING_BACKENDS = {
    'backend_1': {
        'ENGINE': 'eventtracking.backends.routing.RoutingBackend',
        'OPTIONS': {
            'backends': {
                'nested_backend_1': {
                    'ENGINE': 'mock.MagicMock',
                    'OPTIONS': {}
                }
            },
            'processors': [{
                'ENGINE': 'mock.MagicMock',
                'OPTIONS': {}
            }]
        }
    }
}


class TestAsyncSend(TestCase):
    """
    Test `send_event` task.
    """

    def setUp(self):
        super().setUp()

        self.event = {
            'name': str(sentinel.name),
            'event_type': 'edx.test.event',
            'time': '2020-01-01T12:12:12.000000+00:00',
            'event': {
                'key': 'value'
            },
            'session': '0000'
        }

    @override_settings(EVENT_TRACKING_BACKENDS=MOCK_EVENT_TRACKING_BACKENDS)
    def test_successful_processing_and_sending_event(self):
        override_default_tracker()
        tracker = get_tracker()
        tracker.backends['backend_1'].processors[0].return_value = self.event

        processed_event = tracker.backends['backend_1'].process_event(self.event)
        send_event.apply(['backend_1', processed_event])

        tracker.backends['backend_1'].backends['nested_backend_1'].send.assert_called_once_with(self.event)



"""Test the segment.com backend"""


from unittest import TestCase
from unittest.mock import patch
from unittest.mock import sentinel

from eventtracking.backends.segment import SegmentBackend


class TestSegmentBackend(TestCase):
    """Test the segment.com backend"""

    def setUp(self):
        super().setUp()
        patcher = patch('eventtracking.backends.segment.analytics')
        self.addCleanup(patcher.stop)
        self.mock_analytics = patcher.start()
        self.backend = SegmentBackend()

    def test_simple_emit(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id
            },
            'data': {
                'foo': sentinel.bar
            }
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(sentinel.user_id, sentinel.name, event, context={})

    def test_missing_name(self):
        event = {}
        self.backend.send(event)
        self.assert_no_event_emitted()

    def assert_no_event_emitted(self):
        """Ensure no event was actually sent to segment.com"""
        self.assertEqual(len(self.mock_analytics.mock_calls), 0)

    def test_missing_context(self):
        event = {
            'name': sentinel.name,
        }
        self.backend.send(event)
        self.assert_no_event_emitted()

    def test_missing_user_id(self):
        event = {
            'name': sentinel.name,
            'context': {}
        }
        self.backend.send(event)
        self.assert_no_event_emitted()

    def test_google_analytics_client_id(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id,
                'client_id': sentinel.client_id
            }
        }
        expected_segment_context = {
            'Google Analytics': {
                'clientId': sentinel.client_id
            }
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(
            sentinel.user_id, sentinel.name, event, context=expected_segment_context)

    def test_ip_address(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id,
                'ip': sentinel.ip
            }
        }
        expected_segment_context = {
            'ip': sentinel.ip
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(
            sentinel.user_id, sentinel.name, event, context=expected_segment_context)

    def test_user_agent(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id,
                'agent': sentinel.user_agent
            }
        }
        expected_segment_context = {
            'userAgent': sentinel.user_agent
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(
            sentinel.user_id, sentinel.name, event, context=expected_segment_context)

    def test_path(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id,
                'path': sentinel.path
            }
        }
        expected_segment_context = {
            'page': {
                'path': sentinel.path
            }
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(
            sentinel.user_id, sentinel.name, event, context=expected_segment_context)

    def test_referer(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id,
                'referer': sentinel.referer
            }
        }
        expected_segment_context = {
            'page': {
                'referrer': sentinel.referer
            }
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(
            sentinel.user_id, sentinel.name, event, context=expected_segment_context)

    def test_page(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id,
                'page': sentinel.page
            }
        }
        expected_segment_context = {
            'page': {
                'url': sentinel.page
            }
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(
            sentinel.user_id, sentinel.name, event, context=expected_segment_context)

    def test_host_and_path_with_missing_page(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id,
                # Note that 'host' and 'path' will be urlparsed, so must be strings.
                'path': '/this/is/a/path',
                'host': 'hostname',
            }
        }
        expected_segment_context = {
            'page': {
                'path': '/this/is/a/path',
                'url': 'https://hostname/this/is/a/path'  # Synthesized URL value.
            }
        }
        self.backend.send(event)
        self.mock_analytics.track.assert_called_once_with(
            sentinel.user_id, sentinel.name, event, context=expected_segment_context)


class TestSegmentBackendMissingDependency(TestCase):
    """Test the segment.com backend without the package installed"""

    def test_no_analytics_api(self):
        event = {
            'name': sentinel.name,
            'context': {
                'user_id': sentinel.user_id
            },
            'data': {
                'foo': sentinel.bar
            }
        }
        backend = SegmentBackend()
        backend.send(event)



"""
Test the async routing backend.
"""
from unittest import TestCase

from unittest.mock import sentinel, patch
from eventtracking.backends.async_routing import AsyncRoutingBackend


class TestAsyncRoutingBackend(TestCase):
    """
    Test the async routing backend.
    """

    def setUp(self):
        super().setUp()
        self.sample_event = {
            'name': str(sentinel.name),
            'event_type': 'edx.test.event',
            'time': '2020-01-01T12:12:12.000000+00:00',
            'event': {
                'key': 'value'
            },
            'session': '0000'
        }

    @patch('eventtracking.backends.async_routing.send_event')
    def test_successful_event_send(self, mocked_send_event):
        backend = AsyncRoutingBackend(backend_name='test')
        processed_event = backend.process_event(self.sample_event)
        backend.send(self.sample_event)
        mocked_send_event.delay.assert_called_once_with('test', processed_event)



"""Test the routing backend"""


from unittest import TestCase

from unittest.mock import MagicMock, sentinel

from eventtracking.backends.routing import RoutingBackend
from eventtracking.processors.exceptions import EventEmissionExit


class TestRoutingBackend(TestCase):
    """Test the routing backend"""

    def setUp(self):
        super().setUp()
        self.sample_event = {'name': sentinel.name}

        self.mock_backend = MagicMock()
        self.router = RoutingBackend(backends={'0': self.mock_backend})

    def test_non_callable_backend(self):
        with self.assertRaisesRegexp(  # pylint: disable=deprecated-method,useless-suppression
                ValueError, r'Backend \w+ does not have a callable "send" method.'):
            RoutingBackend(backends={
                'a': 'b'
            })

    def test_backend_without_send(self):
        with self.assertRaisesRegexp(  # pylint: disable=deprecated-method,useless-suppression
                ValueError, r'Backend \w+ does not have a callable "send" method.'):
            RoutingBackend(backends={
                'a': object()
            })

    def test_non_callable_processor(self):
        with self.assertRaisesRegexp(  # pylint: disable=deprecated-method,useless-suppression
                ValueError, r'Processor \w+ is not callable.'):
            RoutingBackend(processors=[
                object()
            ])

    def test_non_callable_processor_simple_type(self):
        with self.assertRaisesRegexp(  # pylint: disable=deprecated-method,useless-suppression
                ValueError, r'Processor \w+ is not callable.'):
            RoutingBackend(processors=[
                'b'
            ])

    def test_single_processor(self):
        mock_processor = MagicMock()
        router = RoutingBackend(processors=[
            mock_processor
        ])
        router.send(self.sample_event)
        mock_processor.assert_called_once_with(self.sample_event)

    def test_single_backend(self):
        mock_backend = MagicMock()
        router = RoutingBackend(backends={
            'mock0': mock_backend
        })
        router.send(self.sample_event)
        mock_backend.send.assert_called_once_with(self.sample_event)

    def test_multiple_backends(self):
        backends = {
            str(i): MagicMock()
            for i in range(5)
        }
        router = RoutingBackend(backends=backends)
        router.send(self.sample_event)
        for backend in backends.values():
            backend.send.assert_called_once_with(self.sample_event)

    def test_backend_failure(self):
        backends = {
            str(i): MagicMock()
            for i in range(5)
        }
        backends['1'].send.side_effect = RuntimeError
        router = RoutingBackend(backends=backends)
        router.send(self.sample_event)
        for backend in backends.values():
            backend.send.assert_called_once_with(self.sample_event)

    def test_multiple_processors(self):
        processors = [
            MagicMock()
            for __ in range(5)
        ]
        for processor in processors:
            processor.return_value = self.sample_event
        router = RoutingBackend(processors=processors)
        router.send(self.sample_event)
        for processor in processors:
            processor.assert_called_once_with(self.sample_event)

    def test_multiple_backends_and_processors(self):
        backends = {
            str(i): MagicMock()
            for i in range(5)
        }
        processors = [
            MagicMock()
            for __ in range(5)
        ]
        for processor in processors:
            processor.return_value = self.sample_event

        router = RoutingBackend(backends=backends, processors=processors)
        router.send(self.sample_event)
        for processor in processors:
            processor.assert_called_once_with(self.sample_event)
        for backend in backends.values():
            backend.send.assert_called_once_with(self.sample_event)

    def test_callable_class_processor(self):
        class SampleProcessor:
            """An event processing class"""

            def __call__(self, event):
                """Modify the event type"""
                event['name'] = sentinel.changed_name

        self.router.register_processor(SampleProcessor())
        self.router.send(self.sample_event)
        self.assert_single_event_emitted({'name': sentinel.changed_name})

    def assert_single_event_emitted(self, event):
        """Assert that the mock backend is called exactly once with the provided event"""
        self.mock_backend.send.assert_called_once_with(event)

    def test_function_processor(self):
        def change_name(event):
            """Modify the event type of the event"""
            event['name'] = sentinel.changed_name
            return event

        self.router.register_processor(change_name)
        self.router.send(self.sample_event)
        self.assert_single_event_emitted({'name': sentinel.changed_name})

    def test_processor_chain(self):

        def change_name(event):
            """Modify the event type of the event"""
            event['name'] = sentinel.changed_name
            return event

        def inject_fields(event):
            """Add a couple fields to the event"""
            event['other'] = sentinel.other
            event['to_remove'] = sentinel.to_remove
            return event

        def remove_field(event):
            """Remove a field to the event"""
            self.assertEqual(event['to_remove'], sentinel.to_remove)
            del event['to_remove']
            return event

        def ensure_modified_event(event):
            """Assert the first processor added a field to the event"""
            self.assertEqual(event['name'], sentinel.changed_name)
            self.assertEqual(event['other'], sentinel.other)
            return event

        self.router.register_processor(change_name)
        self.router.register_processor(inject_fields)
        self.router.register_processor(remove_field)
        self.router.register_processor(ensure_modified_event)
        self.router.send(self.sample_event)

        self.assert_single_event_emitted(
            {
                'name': sentinel.changed_name,
                'other': sentinel.other
            }
        )

    def test_processor_failure(self):

        def always_fail(event):  # pylint: disable=unused-argument, useless-suppression
            """Always raises an error"""
            raise ValueError

        def change_name(event):
            """Modify the event type of the event"""
            event['name'] = sentinel.changed_name
            return event

        self.router.register_processor(always_fail)
        self.router.register_processor(change_name)
        self.router.send(self.sample_event)
        self.assert_single_event_emitted({'name': sentinel.changed_name})

    def test_processor_returns_none(self):

        def return_none(event):  # pylint: disable=unused-argument
            """Don't return the event"""

        self.router.register_processor(return_none)
        self.router.send(self.sample_event)
        self.assert_single_event_emitted(self.sample_event)

    def test_processor_modifies_the_same_event_object(self):

        def forget_return(event):
            """Modify the event without returning it"""
            event['name'] = sentinel.forgotten_return

        def ensure_name_changed(event):
            """Assert the event type has been modified even though the event wasn't returned"""
            self.assertEqual(event['name'], sentinel.forgotten_return)

        self.router.register_processor(forget_return)
        self.router.register_processor(ensure_name_changed)
        self.router.send(self.sample_event)
        self.assert_single_event_emitted({'name': sentinel.forgotten_return})

    def test_processor_abort(self):

        def abort_processing(event):  # pylint: disable=unused-argument, useless-suppression
            """Always abort processing"""
            raise EventEmissionExit

        def fail_if_called(event):  # pylint: disable=unused-argument
            """Fail the test immediately if this is called"""
            self.fail('This processor should never be called')

        self.router.register_processor(abort_processing)
        self.router.register_processor(fail_if_called)
        self.router.send(self.sample_event)
        self.assertEqual(len(self.mock_backend.mock_calls), 0)

    def test_nested_routing_with_abort(self):
        mock_abort_processing = MagicMock()
        mock_abort_processing.side_effect = EventEmissionExit

        left_backend = MagicMock()
        left_router = RoutingBackend(backends={'0': left_backend}, processors=[mock_abort_processing])

        right_backend = MagicMock()
        right_router = RoutingBackend(backends={'0': right_backend})

        root_router = RoutingBackend(backends={
            'left': left_router,
            'right': right_router,
        })

        root_router.send(self.sample_event)

        right_backend.send.assert_called_once_with(self.sample_event)
        self.assertEqual(len(left_backend.mock_calls), 0)
        mock_abort_processing.assert_called_once_with(self.sample_event)

    def test_backend_call_order(self):

        class OrderRecordingBackend:
            """Keep track of the order that the backends are called in"""

            def __init__(self, name, call_order):
                self._name = name
                self._order = call_order

            def send(self, event):  # pylint: disable=unused-argument
                """Do nothing except record that this was called"""
                self._order.append(self._name)

        call_order = []
        backends = {
            str(i): OrderRecordingBackend(str(i), call_order)
            for i in range(5)
        }

        router = RoutingBackend(backends=backends)

        router.send(self.sample_event)
        self.assertEqual(call_order, ['0', '1', '2', '3', '4'])



"""Test the whitelist processor"""


from unittest import TestCase

from unittest.mock import sentinel

from eventtracking.processors.exceptions import EventEmissionExit
from eventtracking.processors.whitelist import NameWhitelistProcessor


class TestNameWhitelistProcessor(TestCase):
    """Test the whitelist processor"""

    def test_filtering_out(self):
        whitelist = NameWhitelistProcessor(whitelist=[sentinel.allowed_event])
        with self.assertRaises(EventEmissionExit):
            whitelist({'name': sentinel.not_allowed_event})

    def test_allowed_event(self):
        whitelist = NameWhitelistProcessor(whitelist=[sentinel.allowed_event])
        self.assert_event_passed_through(whitelist, {'name': sentinel.allowed_event})

    def assert_event_passed_through(self, whitelist, event):
        """Assert that the whitelist allowed the event processing to proceed"""
        self.assertEqual(whitelist(event), event)

    def test_empty_whitelist(self):
        whitelist = NameWhitelistProcessor(whitelist=[])
        with self.assertRaises(EventEmissionExit):
            whitelist({'name': sentinel.not_allowed_event})

    def test_multi_entry_whitelist(self):
        whitelist = NameWhitelistProcessor(whitelist=[sentinel.allowed_event, sentinel.another_event])
        with self.assertRaises(EventEmissionExit):
            whitelist({'name': sentinel.not_allowed_event})

        self.assert_event_passed_through(whitelist, {'name': sentinel.allowed_event})
        self.assert_event_passed_through(whitelist, {'name': sentinel.another_event})

    def test_no_whitelist_param(self):
        with self.assert_initialization_fails():
            NameWhitelistProcessor()

    def assert_initialization_fails(self):
        """Assert that the constructor raises the expected initialization exception"""
        return self.assertRaisesRegexp(  # pylint: disable=deprecated-method,useless-suppression
            TypeError, r'The NameWhitelistProcessor must be passed')

    def test_whitelist_param_not_iterable(self):
        with self.assert_initialization_fails():
            NameWhitelistProcessor(10)

    def test_whitelist_param_just_a_string(self):
        with self.assert_initialization_fails():
            NameWhitelistProcessor('foobar')

    def test_whitelist_param_is_none(self):
        with self.assert_initialization_fails():
            NameWhitelistProcessor(None)

    def test_initialize_with_set(self):
        self.assert_properly_configured(frozenset([sentinel.allowed_event]))

    def assert_properly_configured(self, allowed_names):
        """Assert that whitelist was configured properly by correctly passing and/or filtering events"""
        whitelist = NameWhitelistProcessor(whitelist=allowed_names)
        self.assert_event_passed_through(whitelist, {'name': sentinel.allowed_event})
        with self.assertRaises(EventEmissionExit):
            whitelist({'name': sentinel.not_allowed_event})

    def test_initialize_with_dict(self):
        self.assert_properly_configured({sentinel.allowed_event: sentinel.discarded})



"""
Test the event tracking module
"""


from datetime import datetime
from unittest import TestCase


from eventtracking import tracker
from unittest.mock import MagicMock, call, patch, sentinel  # pylint: disable=wrong-import-order
from pytz import UTC  # pylint: disable=wrong-import-order


class TestTrack(TestCase):
    """Tests cases for the event tracking module."""

    def setUp(self):
        super().setUp()
        self._mock_backend = None
        self._mock_backends = []
        self.tracker = None
        self.configure_mock_backends(1)

        self._expected_timestamp = datetime.now(UTC)
        self._datetime_patcher = patch('eventtracking.tracker.datetime')
        self.addCleanup(self._datetime_patcher.stop)
        mock_datetime = self._datetime_patcher.start()
        mock_datetime.now.return_value = self._expected_timestamp

    def configure_mock_backends(self, number_of_mocks):
        """Ensure the tracking module has the requisite number of mock backends"""
        backends = {}
        for i in range(number_of_mocks):
            name = f'mock{i}'
            backend = MagicMock()
            backends[name] = backend

        self.tracker = tracker.Tracker(backends)
        tracker.register_tracker(self.tracker)
        self._mock_backends = list(backends.values())
        self._mock_backend = self._mock_backends[0]

    def get_mock_backend(self, index):
        """Get the mock backend created by `configure_mock_backends`"""
        return self.tracker.get_backend(f'mock{index}')

    def test_event_simple_event_without_data(self):
        self.tracker.emit(sentinel.name)

        self.assert_backend_called_with(sentinel.name)

    def assert_backend_called_with(self, name, data=None, context=None, backend=None):
        """Ensures the backend is called exactly once with the expected data."""

        self.assert_exact_backend_calls([(name, context, data)], backend=backend)

    def assert_exact_backend_calls(self, parameter_tuple_list, backend=None):
        """
        Ensure the backend was called with the specified parameters in the
        specified order.  Note that it expects a list of tuples to be passed
        in to `parameter_tuple_list`.  Each tuple should be in the form:

        (name, context, data)

        These are expanded out into complete events.
        """
        if not backend:
            backend = self._mock_backend

        self.assertEqual(
            backend.send.mock_calls,
            [
                call({
                    'name': name,
                    'timestamp': self._expected_timestamp,
                    'context': context or {},
                    'data': data or {}
                })
                for name, context, data
                in parameter_tuple_list
            ]
        )

    def test_event_simple_event_without_type(self):
        self.tracker.emit(data={sentinel.key: sentinel.value})

        self.assert_backend_called_with(
            'unknown',
            {
                sentinel.key: sentinel.value
            }
        )

    def test_event_simple_event_without_type_or_data(self):
        self.tracker.emit()
        self.assert_backend_called_with(
            'unknown',
            {}
        )

    def test_event_simple_event_with_data(self):
        self.tracker.emit(
            sentinel.name,
            {
                sentinel.key: sentinel.value
            }
        )

        self.assert_backend_called_with(
            sentinel.name,
            {
                sentinel.key: sentinel.value
            }
        )

    def test_global_tracker(self):
        tracker.emit(sentinel.name)

        self.assert_backend_called_with(
            sentinel.name)

    def test_missing_tracker(self):
        self.assertRaises(KeyError, tracker.get_tracker, 'foobar')

    def test_single_context(self):
        context = {sentinel.context_key: sentinel.context_value}
        data = {sentinel.key: sentinel.value}

        self.tracker.enter_context('single', context)
        self.tracker.emit(sentinel.name, data)
        self.tracker.exit_context('single')

        self.assert_backend_called_with(
            sentinel.name,
            data=data,
            context=context
        )

    def test_context_override(self):
        context = {
            sentinel.context_key: sentinel.context_value,
            sentinel.another_key: sentinel.another_value
        }
        override_context = {
            sentinel.context_key: sentinel.override_context_value
        }
        with self.tracker.context('outer', context):
            with self.tracker.context('inner', override_context):
                self.tracker.emit(sentinel.name)

        self.assert_backend_called_with(
            sentinel.name,
            context={
                sentinel.context_key: sentinel.override_context_value,
                sentinel.another_key: sentinel.another_value
            }
        )

    def test_exception_in_context(self):
        with self.assertRaises(ValueError):
            with self.tracker.context('foo', {sentinel.context_key: sentinel.context_value}):
                raise ValueError

        self.tracker.emit(sentinel.name)

        self.assert_backend_called_with(sentinel.name)



"""Event tracking backend that sends events to segment.com"""

from six.moves.urllib.parse import urlunsplit

try:
    import analytics
except ImportError:
    analytics = None


class SegmentBackend:
    """

    Send events to segment.com

    It is assumed that other code elsewhere initializes the segment.com API and makes calls to analytics.identify.

    Requires all emitted events to have the following structure (at a minimum)::

        {
            'name': 'something',
            'context': {
                'user_id': 10,
            }
        }

    Additionally, the following fields can optionally be defined::

        {
            'context': {
                'agent': "your user-agent string",
                'client_id': "your google analytics client id",
                'host': "your hostname",
                'ip': "your IP address",
                'page': "your page",
                'path': "your path",
                'referer': "your referrer",
            }
        }

    The 'page', 'path' and 'referer' are sent to Segment as "page" information.  If the 'page' is absent but the 'host'
    and 'path' are present, these are used to create a URL value to substitute for the 'page' value.

    Note that although some parts of the event are lifted out to pass explicitly into the Segment.com API, the entire
    event is sent as the payload to segment.com, which includes all context, data and other fields in the event.

    """

    def send(self, event):
        """Use the segment.com python API to send the event to segment.com"""
        if analytics is None:
            return

        context = event.get('context', {})
        user_id = context.get('user_id')
        name = event.get('name')
        if name is None or user_id is None:
            return

        segment_context = {}

        ga_client_id = context.get('client_id')
        if ga_client_id is not None:
            segment_context['Google Analytics'] = {
                'clientId': ga_client_id
            }
        ip_address = context.get('ip')
        if ip_address is not None:
            segment_context['ip'] = ip_address
        user_agent = context.get('agent')
        if user_agent is not None:
            segment_context['userAgent'] = user_agent
        path = context.get('path')
        referer = context.get('referer')
        page = context.get('page')

        if path and not page:
            # Try to put together a url from host and path, hardcoding the schema.
            # (Segment doesn't care about the schema for GA, but will extract the host and path from the url.)
            host = context.get('host')
            if host:
                parts = ("https", host, path, "", "")
                page = urlunsplit(parts)

        if path is not None or referer is not None or page is not None:
            segment_context['page'] = {}
            if path is not None:
                segment_context['page']['path'] = path
            if referer is not None:
                segment_context['page']['referrer'] = referer
            if page is not None:
                segment_context['page']['url'] = page

        analytics.track(
            user_id,
            name,
            event,
            context=segment_context
        )



"""
New AppConfig for Django 1.8
"""
from django.apps import AppConfig


class EventTrackingConfig(AppConfig):
    """
    Django 1.8 requires unique app labels and only uses the characters to the right of the
    last period in the string. .django was not specific enough.
    """

    name = 'eventtracking.django'
    label = 'eventtracking_django'

    def ready(self):
        """
        Initialize django specific tracker.
        """
        super().ready()
        # pylint: disable=import-outside-toplevel
        from eventtracking.django.django_tracker import override_default_tracker
        override_default_tracker()



"""Route events to processors and backends"""


import logging
from collections import OrderedDict
from copy import deepcopy

from eventtracking.processors.exceptions import (
    EventEmissionExit,
    NoBackendEnabled,
    NoTransformerImplemented,
)


LOG = logging.getLogger(__name__)


class RoutingBackend:
    """

    Route events to the appropriate backends.

    A routing backend has two types of components:

    1) Processors - These are run sequentially, processing the output of the previous processor. If you had three
       processors [a, b, c], the output of the processing step would be `c(b(a(event)))`. Note that for performance
       reasons, the processor is able to actually mutate the event dictionary in-place. Event dictionaries may be large
       and highly nested, so creating multiple copies could be problematic. A processor can also choose to prevent the
       event from being emitted by raising `EventEmissionExit`. Doing so will prevent any subsequent processors from
       running and prevent the event from being sent to the backends. Any other exception raised by a processor will be
       logged and swallowed, subsequent processors will execute and the event will be emitted.
    2) Backends - Backends are intended to not mutate the event and each receive the same event data. They are not
       chained like processors. Once an event has been processed by the processor chain, it is passed to each backend in
       the order that they were registered. Backends typically persist the event in some way, either by sending it
       to an external system or saving it to disk. They are called synchronously and in sequence, so a long running
       backend will block other backends until it is done persisting the event. Note that you can register another
       `RoutingBackend` as a backend of a `RoutingBackend`, allowing for arbitrary processing trees.

    `backends` is a collection that supports iteration over its items using `iteritems()`. The keys are expected to be
        sortable and the values are expected to expose a `send(event)` method that will be called for each event. Each
        backend in this collection is registered in order sorted alphanumeric ascending by key.
    `processors` is an iterable of callables.

    Raises a `ValueError` if any of the provided backends do not have a callable "send" attribute or any of the
        processors are not callable.
    """

    def __init__(self, backends=None, processors=None):
        self.backends = OrderedDict()
        self.processors = []

        if backends is not None:
            for name in sorted(backends.keys()):
                self.register_backend(name, backends[name])

        if processors is not None:
            for processor in processors:
                self.register_processor(processor)

    def register_backend(self, name, backend):
        """
        Register a new backend that will be called for each processed event.

        Note that backends are called in the order that they are registered.
        """
        if not hasattr(backend, 'send') or not callable(backend.send):
            raise ValueError('Backend %s does not have a callable "send" method.' % backend.__class__.__name__)

        self.backends[name] = backend

    def register_processor(self, processor):
        """
        Register a new processor.

        Note that processors are called in the order that they are registered.
        """
        if not callable(processor):
            raise ValueError('Processor %s is not callable.' % processor.__class__.__name__)

        self.processors.append(processor)

    def send(self, event):
        """
        Process the event using all registered processors and send it to all registered backends.

        Logs and swallows all `Exception`.
        """
        event = deepcopy(event)

        try:
            processed_event = self.process_event(event)
        except EventEmissionExit:
            return
        else:
            self.send_to_backends(processed_event)

    def process_event(self, event):
        """

        Executes all event processors on the event in order.

        `event` is a nested dictionary that represents the event.

        Logs and swallows all `Exception` except `EventEmissionExit` which is re-raised if it is raised by a processor.

        Returns the modified event.
        """

        if len(self.processors) == 0:
            return event

        processed_event = event

        for processor in self.processors:
            try:
                modified_event = processor(processed_event)
                if modified_event is not None:
                    processed_event = modified_event
            except EventEmissionExit:
                raise
            except Exception:  # pylint: disable=broad-except
                LOG.exception(
                    'Failed to execute processor: %s', str(processor)
                )

        return processed_event

    def send_to_backends(self, event):
        """
        Sends the event to all registered backends.

        Logs and swallows all `Exception`.
        """

        for name, backend in self.backends.items():
            try:
                backend.send(event)
            except NoTransformerImplemented as exc:
                LOG.info('[send_to_backends] No transformer has been implemented for edx event "%s", [%s]',
                         event["name"], repr(exc))
            except NoBackendEnabled as exc:
                LOG.info('[send_to_backends] Failed to send edx event "%s" to "%s" backend. "%s" backend has'
                         ' not been enabled, [%s]', event["name"], name, name, repr(exc)
                         )
            except Exception:  # pylint: disable=broad-except
                LOG.exception(
                    'Unable to send edx event "%s" to backend: %s', event["name"], name
                )



"""
Celery tasks
"""

from celery.utils.log import get_task_logger
from celery import shared_task
from edx_django_utils.monitoring import set_code_owner_attribute_from_module
from eventtracking.tracker import get_tracker
from eventtracking.processors.exceptions import (
    NoBackendEnabled,
    NoTransformerImplemented,
)

logger = get_task_logger(__name__)
# Maximum number of retries before giving up on rounting event
MAX_RETRIES = 3
# Number of seconds after task is retried
COUNTDOWN = 30


@shared_task(bind=True)
def send_event(self, backend_name, processed_event):
    """
    Send event to configured top-level backend asynchronously.

    WARNING: Do not use this task directly! It is intended for use
    only by the AsyncRoutingBackend, since it is implemented with the
    following assumptions:

    - That the top-level processors have already been run on the event
    - That the named backend is a RoutingBackend (or descendent)

    Arguments:
        self (dict): task
        backend_name (str):  name of the backend to use
        processed_event (dict): Processed event dict
    """
    set_code_owner_attribute_from_module(self.__module__)
    try:
        tracker = get_tracker()
        backend = tracker.backends[backend_name]
        backend.send_to_backends(processed_event.copy())

    except (NoTransformerImplemented, NoBackendEnabled) as exc:
        logger.info(
            '[send_event] Failed to send event [%s] with backend [%s], [%s]',
            processed_event['name'], backend_name, exc
        )

    except Exception as exc:
        logger.exception(
            '[send_event] Failed to send event [%s] with backend [%s], [%s]',
            processed_event['name'], backend_name, repr(exc)
        )
        raise self.retry(exc=exc, countdown=COUNTDOWN, max_retries=MAX_RETRIES)



"""
Filter out events by comparing event names with the provided regular expressions.
"""
import re
from logging import getLogger

from eventtracking.processors.exceptions import EventEmissionExit
from eventtracking.exceptions import ImproperlyConfigured


logger = getLogger(__name__)

ALLOWLIST = 'allowlist'
BLOCKLIST = 'blocklist'


class RegexFilter:
    """
    Filter out events by comparing event names with the provided regular expressions.
    """

    def __init__(self, filter_type=ALLOWLIST, regular_expressions=None):
        self.regular_expressions = frozenset(regular_expressions or [])
        self.compiled_expressions = self._compile_regular_expressions()
        self.filter_type = filter_type
        self._validate_filter_type()

    def __call__(self, event):
        is_a_match = self._event_matches_filter(event['name'])

        if (
            (is_a_match and self.filter_type == ALLOWLIST) or
            (not is_a_match and self.filter_type == BLOCKLIST)
        ):
            return event

        raise EventEmissionExit()

    def _validate_filter_type(self):
        """
        Validate that the filter type is either `allowlist` or `blocklist`

        Raises:
            ImproperlyConfigured
        """
        if self.filter_type not in (ALLOWLIST, BLOCKLIST):
            logger.error(
                'Unsupported filter type {} is set. Allowed types are only {} and {}.'.format(
                    self.filter_type,
                    ALLOWLIST,
                    BLOCKLIST
                ))
            raise ImproperlyConfigured('Invalid filter type is configured')

    def _compile_regular_expressions(self):
        """
        Compile and validate every regular expression in the list.
        Throws on invalid regex.

        Returns:
            list<compiled re>
        """
        expressions_list = self.regular_expressions

        invalid_regex_expressions = []
        compiled_regex_expressions = []

        for exp in expressions_list:
            try:
                compiled = re.compile(exp)
                compiled_regex_expressions.append(compiled)
            except re.error:
                invalid_regex_expressions.append(exp)

        if invalid_regex_expressions:
            logger.error('The following invalid regular expressions are configured'
                         'for setting "ASYNC_ROUTING_BACKENDS_FILTERS": {}'.format(invalid_regex_expressions))
            raise ImproperlyConfigured('Invalid Regular Expressions are configured.')

        return compiled_regex_expressions

    def _event_matches_filter(self, event_name):
        """
        Determine if event name matches any of the regular expressions.

        Arguments:
            event_name (str):   name of the event

        Returns:
            bool
        """
        for expression in self.compiled_expressions:
            if expression.match(event_name):
                return True
        return False



"""A Django settings file for testing"""


DEBUG = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:'
    }
}

SITE_ID = 1

MIDDLEWARE = (
    'django.middleware.common.CommonMiddleware',
)

INSTALLED_APPS = [
    'eventtracking.django'
]

EVENT_TRACKING_ENABLED = True

SECRET_KEY = "test_key"



"""
Runs invasive tests to ensure that the tracking system can communicate with
the actual python logging system.
"""


import json
import logging
import os
import tempfile
from datetime import datetime

from pytz import UTC

from eventtracking.backends.logger import LoggerBackend
from eventtracking.backends.tests import InMemoryBackend, IntegrationTestCase
from eventtracking.tracker import Tracker


class TestLoggerIntegration(IntegrationTestCase):
    """
    Writes a real log file and ensures the messages are written properly.
    """

    def setUp(self):

        super().setUp()
        logger_name = 'integration.test'
        test_logger = logging.getLogger(logger_name)
        test_logger.setLevel(logging.INFO)

        self.temporary_fd, self.temporary_file_name = tempfile.mkstemp()
        self.addCleanup(os.remove, self.temporary_file_name)

        self.temporary_file_handler = logging.FileHandler(self.temporary_file_name, mode='w', encoding='utf_8')
        self.temporary_file_handler.setFormatter(logging.Formatter(fmt='%(message)s'))
        test_logger.addHandler(self.temporary_file_handler)
        self.addCleanup(test_logger.removeHandler, self.temporary_file_handler)

        self.logger_backend = LoggerBackend(name=logger_name)
        self.memory_backend = InMemoryBackend()
        self.tracker = Tracker({
            'logger': self.logger_backend,
            'mem': self.memory_backend
        })

    def test_sequential_events(self):
        now = datetime.now(UTC)
        for i in range(10):
            self.tracker.emit('org.test.logger.integration', {
                'email': 'tester@eventtracking.org',
                'sequence': i,
                'current_time': now
            })

        # The custom JSON encoder will string encode these fields, however
        # it is not used to decode the events, so we need to compare these
        # fields as strings.
        for event in self.memory_backend.events:
            event['timestamp'] = event['timestamp'].isoformat()
            event['data']['current_time'] = event['data']['current_time'].isoformat()
        self.assertEqual(len(self.memory_backend.events), 10)

        written_events = []
        with os.fdopen(self.temporary_fd, 'r') as temporary_file:
            for line in temporary_file:
                loaded_event = json.loads(line.strip())
                written_events.append(loaded_event)

        self.assertEqual(written_events, self.memory_backend.events)



"""
Package metadata for event-tracking.
"""

import os
from setuptools import setup
from setuptools import find_packages

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement;
    that is, it is not blank, a comment, or editable.
    """
    # Remove whitespace at the start/end of the line
    line = line.strip()

    # Skip blank lines, comments, and editable installs
    return not (
        line == '' or
        line.startswith('-c') or
        line.startswith('-r') or
        line.startswith('#') or
        line.startswith('-e') or
        line.startswith('git+')
    )


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.
    Returns a list of requirement strings.
    """
    requirements = set()
    for path in requirements_paths:
        requirements.update(
            line.strip() for line in open(path, encoding='utf-8').readlines()
            if is_requirement(line)
        )
    return list(requirements)


README = open(os.path.join(os.path.dirname(__file__), 'README.rst'),
              encoding='utf-8').read()
REQUIREMENTS = load_requirements('requirements/base.in')

setup(
    name='event-tracking',
    version='1.1.4',
    packages=find_packages(),
    include_package_data=True,
    license='AGPLv3 License',
    description='A simple event tracking system.',
    long_description=README,
    install_requires=REQUIREMENTS,
    url='https://github.com/edx/event-tracking',
    author='edX',
    author_email='oscm@edx.org',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



"""
Route events to processors and backends.
"""
import logging

from eventtracking.backends.routing import RoutingBackend
from eventtracking.tasks import send_event
from eventtracking.processors.exceptions import EventEmissionExit

logger = logging.getLogger(__name__)


class AsyncRoutingBackend(RoutingBackend):
    """
    Route events to configured backends asynchronously.

    Load the backend with name `backend_name` and use it to process and send
    the event to configured nested backends.

    NB: This can only be safely configured as a top-level backend,
    since the Celery task has to look up the backend again by name.
    This backend can also only be used from the default tracker, since
    again the Celery task does not know which other tracker to use.
    """
    def __init__(self, processors=None, backends=None, backend_name=''):
        self.backend_name = backend_name
        super().__init__(processors=processors, backends=backends)

    def send(self, event):
        """
        Send event to registered backends asynchronously.

        Arguments:
            event (dict) :  Open edX generated analytics event
        """
        try:
            processed_event = self.process_event(event)
            logger.info('Successfully processed event "{}"'.format(event['name']))

        except EventEmissionExit:
            logger.info('[EventEmissionExit] skipping event {}'.format(event['name']))
            return
        send_event.delay(self.backend_name, processed_event)
        logger.info('Scheduled celery task for event "{}" processing and routing'.format(event['name']))



"""
A django specific tracker.
"""
from importlib import import_module

from django.conf import settings

from eventtracking import tracker
from eventtracking.tracker import Tracker
from eventtracking.locator import ThreadLocalContextLocator


DJANGO_BACKEND_SETTING_NAME = 'EVENT_TRACKING_BACKENDS'
DJANGO_PROCESSOR_SETTING_NAME = 'EVENT_TRACKING_PROCESSORS'
DJANGO_ENABLED_SETTING_NAME = 'EVENT_TRACKING_ENABLED'


class DjangoTracker(Tracker):
    """
    A `eventtracking.tracker.Tracker` that constructs its backends from
    Django settings.
    """

    def __init__(
            self,
            backends_settings_name=DJANGO_BACKEND_SETTING_NAME,
            processors_settings_name=DJANGO_PROCESSOR_SETTING_NAME
    ):
        backends = self.create_backends_from_settings(backends_settings_name)
        processors = self.create_processors_from_settings(processors_settings_name)
        super().__init__(backends, ThreadLocalContextLocator(), processors)

    def create_backends_from_settings(self, settings_name):
        """
        Expects the Django setting with `settings_name` (default: "EVENT_TRACKING_BACKENDS")
        to be defined and point to a dictionary of backend engine configurations.

        Example::

            EVENT_TRACKING_BACKENDS = {
                'default': {
                    'ENGINE': 'some.arbitrary.Backend',
                    'OPTIONS': {
                        'endpoint': 'http://something/event'
                    }
                },
                'another_engine': {
                    'ENGINE': 'some.arbitrary.OtherBackend',
                    'OPTIONS': {
                        'user': 'foo'
                    }
                },
            }
        """
        config = getattr(settings, settings_name, {})

        backends = self._instantiate_objects(config)

        return backends

    def _instantiate_objects(self, node):
        """
        Recursively traverse a structure to identify dictionaries that represent objects that need to be instantiated

        Traverse all values of all dictionaries and all elements of all lists to identify dictionaries that contain the
        special "ENGINE" key which indicates that a class of that type should be instantiated and passed all key-value
        pairs found in the sibling "OPTIONS" dictionary as keyword arguments.

        For example::

            tree = {
                'a': {
                    'b': {
                        'first_obj': {
                            'ENGINE': 'mypackage.mymodule.Clazz',
                            'OPTIONS': {
                                'size': 10,
                                'foo': 'bar'
                            }
                        }
                    },
                    'c': [
                        {
                            'ENGINE': 'mypackage.mymodule.Clazz2',
                            'OPTIONS': {
                                'more_objects': {
                                    'd': {'ENGINE': 'mypackage.foo.Bar'}
                                }
                            }
                        }
                    ]
                }
            }
            root = self._instantiate_objects(tree)

        That structure of dicts, lists, and strings will end up with (this example assumes that all keyword arguments to
        constructors were saved as attributes of the same name):

        assert type(root['a']['b']['first_obj']) == <type 'mypackage.mymodule.Clazz'>
        assert root['a']['b']['first_obj'].size == 10
        assert root['a']['b']['first_obj'].foo == 'bar'
        assert type(root['a']['c'][0]) == <type 'mypackage.mymodule.Clazz2'>
        assert type(root['a']['c'][0].more_objects['d']) == <type 'mypackage.foo.Bar'>
        """
        result = node
        if isinstance(node, dict):
            if 'ENGINE' in node:
                result = self._instantiate_from_dict(node)
            else:
                result = {}
                for key, value in node.items():
                    result[key] = self._instantiate_objects(value)
        elif isinstance(node, list):
            result = []
            for child in node:
                result.append(self._instantiate_objects(child))

        return result

    def _instantiate_from_dict(self, values):
        """
        Constructs an object given a dictionary containing an "ENGINE" key
        which contains the full module path to the class, and an "OPTIONS"
        key which contains a dictionary that will be passed in to the
        constructor as keyword args.
        """

        name = values['ENGINE']
        options = values.get('OPTIONS', {})

        # Parse the name
        parts = name.split('.')
        module_name = '.'.join(parts[:-1])
        class_name = parts[-1]

        # Get the class
        try:
            module = import_module(module_name)
            cls = getattr(module, class_name)
        except (ValueError, AttributeError, TypeError, ImportError) as error:
            raise ValueError('Cannot find class %s' % name) from error

        options = self._instantiate_objects(options)
        return cls(**options)

    def create_processors_from_settings(self, settings_name):
        """
        Expects the Django setting with `settings_name` (default: "EVENT_TRACKING_PROCESSORS")
        to be defined and point to a list of backend engine configurations.

        Example::

            EVENT_TRACKING_PROCESSORS = [
                {
                    'ENGINE': 'some.arbitrary.Processor'
                },
                {
                    'ENGINE': 'some.arbitrary.OtherProcessor',
                    'OPTIONS': {
                        'user': 'foo'
                    }
                },
            ]
        """
        config = getattr(settings, settings_name, [])

        processors = self._instantiate_objects(config)

        return processors


def override_default_tracker():
    """Sets the default tracker to a DjangoTracker"""
    if getattr(settings, DJANGO_ENABLED_SETTING_NAME, False):
        tracker.register_tracker(DjangoTracker())



"""Test the logging backend"""


import json
import datetime
from unittest import TestCase

from unittest.mock import patch
from unittest.mock import sentinel
import pytz

from eventtracking.backends.logger import LoggerBackend


class TestLoggerBackend(TestCase):
    """Test the logging backend"""

    def setUp(self):
        super().setUp()
        patcher = patch('eventtracking.backends.logger.logging')
        self.mock_logging = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_logger = self.mock_logging.getLogger.return_value

        # This will call the mocks
        self.backend = LoggerBackend()

        # Reset them so that we get "clean" mocks in our tests
        self.mock_logging.reset_mock()
        self.mock_logger.reset_mock()

    def test_logs_to_correct_named_logger(self):
        backend = LoggerBackend(name=sentinel.logger_name)
        self.mock_logging.getLogger.assert_called_once_with(sentinel.logger_name)
        backend.send({})
        self.assert_event_emitted({})

    def assert_event_emitted(self, event):
        """Asserts exactly one event was emitted"""
        self.mock_logger.info.assert_called_once_with(
            json.dumps(event)
        )

    def test_unserializable_event(self):
        with self.assertRaises(TypeError):
            self.backend.send({'foo': object()})
        self.assert_no_events_emitted()

    def assert_no_events_emitted(self):
        """Asserts no events have been emitted"""
        self.assertFalse(self.mock_logger.info.called)

    def test_big_event(self):
        backend = LoggerBackend(max_event_size=10)
        backend.send({'foo': 'a'*(backend.max_event_size + 1)})
        self.assert_no_events_emitted()

    def test_unlimited_event_size(self):
        default_max_event_size = self.backend.max_event_size
        backend = LoggerBackend(max_event_size=None)
        event = {'foo': 'a'*(default_max_event_size + 1)}
        backend.send(event)
        self.assert_event_emitted(event)

    def test_event_with_datetime_fields(self):
        eastern_tz = pytz.timezone('US/Eastern')
        test_time = datetime.datetime(2012, 5, 1, 7, 27, 1, 200)
        event = {
            'test': True,
            'time': test_time,
            'converted_time': eastern_tz.localize(test_time),
            'date': datetime.date(2012, 5, 7)
        }

        self.backend.send(event)
        self.assert_event_emitted({
            'test': True,
            'time': '2012-05-01T07:27:01.000200+00:00',
            'converted_time': '2012-05-01T11:27:01.000200+00:00',
            'date': '2012-05-07'
        })

    def test_multiple_events(self):
        for event in [{'a': 'a'}, {'b': 'b'}]:
            self.backend.send(event)
            self.assert_event_emitted(event)
            self.mock_logger.info.reset_mock()

    def test_dynamic_level(self):
        backend = LoggerBackend(level='warning')
        backend.send({})
        self.assertFalse(self.mock_logger.info.called)
        self.mock_logger.warning.assert_called_once_with('{}')
